/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER, TCSDESIGNER
 * Version: 2.0
 */

using System;
using System.Collections;
using System.IO;
using System.Web.UI;
using System.Xml;

namespace TopCoder.Web.UI.WebControl.BreadCrumb.PathDiscovery
{
    /// <summary>
    /// This implementation of <see cref="IPathDiscovery"/> and <see cref="ISiteDiscovery"/> will attempt to create a
    /// static site map from either a filename or string source (both of which is an XML document).
    /// This class will parse the document to create a static map of regex matchers (where each node can provide a
    /// regex matching). If used as a <see cref="IPathDiscovery"/>, it provides a static view of the site (i.e. without
    /// dynamic capabilities).
    /// </summary>
    /// <remarks>
    /// This class is thread safe in context of the rendering process. A static session field is used for the site map
    /// after creation, for performance reasons. A thread safety analysis of this fact is present in the component
    /// specification.
    /// </remarks>
    public class XmlPathDiscovery : IPathDiscovery, ISiteDiscovery
    {
        /// <summary>
        /// A "constant" referencing the name of the session variable where the XML site map is stored. Direct access
        /// to it is not recommended, but could for example be used as a way to force the path discovery class to
        /// refresh the map data.
        /// </summary>
        public static readonly string CURRENT_SITEMAP = typeof(XmlPathDiscovery).FullName + ".siteMap";

        /// <summary>
        /// A "constant" referencing the name of the session variable where the XML reversed site map is stored. Direct
        /// access to it is not recommended, but could for example be used as a way to force the path discovery class
        /// to refresh the map data.
        /// </summary>
        public static readonly string REVERSED_SITEMAP = typeof(XmlPathDiscovery).FullName + ".reversedSiteMap";

        /// <summary>
        /// The root element in a valid breadcrumb XML file.
        /// </summary>
        private const string ROOT_ELEMENT = "BreadCrumb";

        /// <summary>
        /// The XML attribute used for patterns.
        /// </summary>
        private const string PATTERN_ATTR = "pattern";

        /// <summary>
        /// The XML attribute used for urls.
        /// </summary>
        private const string URL_ATTR = "url";

        /// <summary>
        /// The XML attribute used for title.
        /// </summary>
        private const string TITLE_ATTR = "title";

        /// <summary>
        /// The schema used when loading. Verification is performed against this schema.
        /// </summary>
        private const string XML_SCHEMA =
            @"<?xml version=""1.0"" encoding=""UTF-8"" ?>
              <xsd:schema
              xmlns:xsd=""http://www.w3.org/2001/XMLSchema"">
                <xsd:element name=""BreadCrumb"" type=""BreadCrumb""/>
                <xsd:complexType name=""BreadCrumb"">
                  <xsd:sequence>
                    <xsd:element name=""bc_node"" type=""bc_node"" minOccurs=""0"" maxOccurs=""unbounded""/>
                  </xsd:sequence>
                  <xsd:attribute name=""version"" type=""xsd:token""/>
                  <xsd:attribute name=""noNamespaceSchemaLocation""
                                 type=""xsd:token""/>
                </xsd:complexType>
                <xsd:complexType name=""bc_node"">
                  <xsd:sequence>
                    <xsd:element maxOccurs=""unbounded"" minOccurs=""0"" name=""bc_node"" type=""bc_node""/>
                  </xsd:sequence>
                  <xsd:attribute name=""pattern"" type=""xsd:token""/>
                  <xsd:attribute name=""title"" type=""xsd:token""/>
                  <xsd:attribute name=""url"" type=""xsd:token""/>
                </xsd:complexType>
              </xsd:schema>";


        /// <summary>
        /// Either the filename of an XML file, or the XML data itself, depending on the setting of
        /// <see cref="isFile"/>.
        /// </summary>
        private string source;

        /// <summary>
        /// Represents whether the source variable represents a file (true) or a raw XML string (false).
        /// </summary>
        private bool isFile;

        /// <summary>
        /// Constructs the <see cref="XmlPathDiscovery"/> from the given filename.
        /// </summary>
        /// <param name="filename">The filename to use.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="filename"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="filename"/> is empty, after trimming.
        /// </exception>
        public XmlPathDiscovery(string filename)
            : this(filename, true)
        {
        }

        /// <summary>
        /// Constructs the <see cref="XmlPathDiscovery"/> from the given source (filename or XML) and parameter
        /// defining the type of source.
        /// </summary>
        /// <param name="source">A non-null, non-empty string being a filename or a direct XML string.</param>
        /// <param name="isFile">A parameter determining whether the source is a file or a XML string.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="source"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="source"/> is empty, after trimming.
        /// </exception>
        public XmlPathDiscovery(string source, bool isFile)
        {
            if (source == null)
            {
                throw new ArgumentNullException("source");
            }
            if (source.Trim().Length == 0)
            {
                throw new ArgumentException("Empty string specified.", "source");
            }

            this.isFile = isFile;
            this.source = source;
        }

        /// <summary>
        /// Reads until an element start, or a file end, is found.
        /// </summary>
        /// <param name="reader">The reader to read from.</param>
        private static void ReadToStartElement(XmlReader reader)
        {
            while (reader.NodeType != XmlNodeType.Element && reader.Read())
            {
            }
        }

        /// <summary>
        /// Reads all bc_node elements from a <see cref="XmlReader"/> and puts them into a tentative site map.
        /// </summary>
        /// <param name="reader">A reader to read from.</param>
        /// <param name="map">The site map to enter entries in.</param>
        /// <exception cref="Exception">Any error.</exception>
        private static void ReadBCNodes(XmlReader reader, IDictionary map)
        {
            ArrayList stack = new ArrayList();
            do
            {
                // Add a node, with an empty list, get on adding some more
                if (reader.NodeType == XmlNodeType.Element)
                {
                    string pattern = reader.GetAttribute(PATTERN_ATTR);
                    string url = reader.GetAttribute(URL_ATTR);
                    string title = reader.GetAttribute(TITLE_ATTR);

                    INodeMatcher nodeMatcher = new NodeMatcher.RegexNodeMatcher(title, url, pattern);
                    IList motherList = null;

                    if (stack.Count > 0)
                    {
                        // A null value means that we have already added children to this node.
                        if (stack[stack.Count - 1] == null)
                        {
                            throw new BreadCrumbException("All bc_node children should be non-null.");
                        }

                        motherList = map[stack[stack.Count - 1]] as IList;
                        motherList.Add(nodeMatcher);
                    }

                    // Only add a list if it doesn't already exist.
                    if (map.Contains(nodeMatcher))
                    {
                        // Get a null element onto the stack, as an error marker, if there are already children
                        // present
                        IList thisList = map[nodeMatcher] as IList;

                        stack.Add(thisList.Count != 0 ? null : nodeMatcher);
                    }
                    else
                    {
                        map[nodeMatcher] = new ArrayList();
                        stack.Add(nodeMatcher);
                    }
                }
                // Pop the stack
                if ((reader.NodeType == XmlNodeType.EndElement || reader.IsEmptyElement) && stack.Count > 0)
                {
                    stack.RemoveAt(stack.Count - 1);
                }
            }
            while ((reader.NodeType != XmlNodeType.EndElement || reader.Name != ROOT_ELEMENT) && reader.Read());
        }

        /// <summary>
        /// Returns a non-null map of <see cref="INodeMatcher"/> instances mapping to <see cref="IList"/> instances
        /// with the matchers of the children.
        /// </summary>
        /// <exception cref="BreadCrumbException">If the parsing fails.</exception>
        /// <returns>A non-null map of <see cref="INodeMatcher"/> instances mapping to <see cref="IList"/> instances
        /// with the <see cref="INodeMatcher">INodeMatchers</see> of the children.</returns>
        public IDictionary GetSiteMap()
        {
            XmlValidatingReader reader = null;
            XmlReader schemaReader = null;
            Stream motherStream = null;

            try
            {
                // Inner try to handle any errors from the closing in "finally" itself
                try
                {
                    if (!isFile)
                    {
                        reader = new XmlValidatingReader(source, XmlNodeType.Document, null);
                    }
                    else
                    {
                        motherStream = File.Open(source, FileMode.Open);
                        reader = new XmlValidatingReader(motherStream, XmlNodeType.Document, null);
                    }

                    schemaReader = new XmlTextReader(new StringReader(XML_SCHEMA));
                    reader.Schemas.Add(null, schemaReader);

                    ReadToStartElement(reader);
                    reader.ReadStartElement(ROOT_ELEMENT);

                    // A variable as an argument is used, since it's quite conceivable that we would like to ADD a
                    // tree to an existing sitemap, not only create a brand new one
                    IDictionary sitemap = new Hashtable();

                    ReadBCNodes(reader, sitemap);

                    return sitemap;
                }
                finally
                {
                    // IDisposable not implemented for this reader in .NET 1.1
                    if (reader != null)
                    {
                        reader.Close();
                    }
                    if (schemaReader != null)
                    {
                        schemaReader.Close();
                    }

                    if (motherStream != null)
                    {
                        motherStream.Close();
                    }
                }
            }
            catch (BreadCrumbException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new BreadCrumbException("Error loading XML.", e);
            }
        }

        /// <summary>
        /// This method attempts to determine the page from the current node to the root and return that path based on
        /// the current site map.
        /// </summary>
        /// <remarks>
        /// The current sitemap is retrieved with the session attribute <see cref="CURRENT_SITEMAP"/>.
        /// </remarks>
        /// <param name="currentNode">A non-null node representing the current page.</param>
        /// <param name="page">The non-null page context.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="currentNode"/> has a null URL or
        /// <paramref name="page"/> is null.</exception>
        /// <exception cref="BreadCrumbException">For any exception while discovering the path.</exception>
        /// <returns>A <see cref="BreadCrumbNode"/> array describing a path from root to current node.</returns>
        public BreadCrumbNode[] GetPath(BreadCrumbNode currentNode, Page page)
        {
            if (currentNode.Url == null)
            {
                throw new ArgumentNullException("currentNode", "BreadCrumbNode not properly initialized.");
            }
            if (page == null)
            {
                throw new ArgumentNullException("page");
            }

            try
            {
                IDictionary sitemap = page.Session[CURRENT_SITEMAP] as IDictionary;
                IDictionary reversedMap = page.Session[REVERSED_SITEMAP] as IDictionary;
                if (sitemap == null || reversedMap == null)
                {
                    sitemap = GetSiteMap();
                    reversedMap = PathDiscoveryHelper.GetReversedMap(sitemap);
                    page.Session[CURRENT_SITEMAP] = sitemap;
                    page.Session[REVERSED_SITEMAP] = reversedMap;
                }

                return PathDiscoveryHelper.SimplifyPath(
                    PathDiscoveryHelper.GetPath(currentNode, null, sitemap, reversedMap));
            }
            catch (BreadCrumbException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new BreadCrumbException("Error retrieving path.", e);
            }
        }
    }
}
