/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER
 * Version: 2.0
 */

using System;
using System.IO;
using System.Web;
using System.Web.UI;
using NUnit.Framework;

namespace TopCoder.Web.UI.WebControl.BreadCrumb
{
    /// <summary>
    /// This class tests the abstract base, intended to be the base class for all path writers, unless a very special
    /// need stops this.
    /// </summary>
    /// <remarks>
    /// We're inheriting from the tested class. It's stateless and we need access to the protected methods.
    /// </remarks>
    [TestFixture]
    public class TestAbstractPathWriter : AbstractPathWriter
    {
        /// <summary>
        /// A series of three test nodes used in some tests.
        /// </summary>
        private BreadCrumbNode[] testNodes;

        /// <summary>
        /// An array of three nodes with null title.
        /// </summary>
        private BreadCrumbNode[] emptyNodes;

        /// <summary>
        /// A node with a null title.
        /// </summary>
        private BreadCrumbNode nullTitleNode;

        /// <summary>
        /// The control used for testing, configured with node styles.
        /// </summary>
        private BreadCrumbTrailControl control;

        /// <summary>
        /// A setting to make <see cref="BeginPath"/> fail.
        /// </summary>
        private bool failBegin = false;

        /// <summary>
        /// A setting to make <see cref="EndPath"/> fail.
        /// </summary>
        private bool failEnd = false;

        /// <summary>
        /// Sets up the test nodes.
        /// </summary>
        [SetUp]
        public void Setup()
        {
            testNodes = new BreadCrumbNode[3];
            testNodes[0] = new BreadCrumbNode("title", "url");
            testNodes[1] = new BreadCrumbNode("title1", "url");
            testNodes[2] = new BreadCrumbNode("title2", "url");

            nullTitleNode = new BreadCrumbNode(null, "url");
            emptyNodes = new BreadCrumbNode[3];
            for (int i = 0; i < 3; i++)
            {
                emptyNodes[i] = nullTitleNode;
            }

            control = new BreadCrumbTrailControl();
            control.NodeStyle = "node";
            control.RootNodeStyle = "root";
            control.MouseOverStyle = "mouse";

            failBegin = false;
            failEnd = false;
        }

        /// <summary>
        /// We just call the base method, BUT we provide an option to fail.
        /// </summary>
        /// <param name="writer">The writer.</param>
        protected override void BeginPath(HtmlTextWriter writer)
        {
            if (failBegin)
            {
                throw new Exception();
            }
            base.BeginPath(writer);
        }

        /// <summary>
        /// We just call the base method, BUT we provide an option to fail.
        /// </summary>
        /// <param name="writer">The writer.</param>
        protected override void EndPath(HtmlTextWriter writer)
        {
            if (failEnd)
            {
                throw new Exception();
            }
            base.EndPath(writer);
        }

        /// <summary>
        /// Tests the path starting method, which does nothing in this class.
        /// </summary>
        [Test]
        public void TestBeginPath()
        {
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    BeginPath(htmlWriter);
                }

                Assert.AreEqual("", writer.ToString(), "Nothing should be written.");
            }
        }

        /// <summary>
        /// Tests the path ending method, which should do nothing in this base class.
        /// </summary>
        [Test]
        public void TestEndPath()
        {
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    EndPath(htmlWriter);
                }

                Assert.AreEqual("", writer.ToString(), "Nothing should be written.");
            }
        }

        /// <summary>
        /// Tests separator writing, which should be an encoded version of the string.
        /// </summary>
        [Test]
        public void TestWriteSeparator()
        {
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    BreadCrumbTrailControl control = new BreadCrumbTrailControl();
                    control.PathSeparator = "&";
                    control.PathSeparatorStyle = "STAJL";
                    WriteSeparator(htmlWriter, control);
                }

                Assert.AreEqual("<span class=\"STAJL\">&amp;</span>", writer.ToString(),
                    "The separator should be encoded and written.");
            }
        }

        /// <summary>
        /// A quick implementation of the node writer itself.
        /// </summary>
        /// <param name="writer">The writer to use.</param>
        /// <param name="node">The node to write.</param>
        /// <param name="nodeStyle">Desired node style.</param>
        /// <param name="mouseOverStyle">Desired mouse over style.</param>
        protected override void WriteNode(HtmlTextWriter writer, BreadCrumbNode node, string nodeStyle,
            string mouseOverStyle)
        {
            // The point of this length check is realy to make sure we throw on null values.
            if (node.Title.Length != 0)
            {
                writer.Write(HttpUtility.HtmlEncode(node.Title));
                writer.Write(HttpUtility.HtmlEncode(nodeStyle));
                writer.Write(HttpUtility.HtmlEncode(mouseOverStyle));
            }
        }

        /// <summary>
        /// Tests writing three entries, in a "forward" direction (RootToCurrent).
        /// </summary>
        [Test]
        public void TestWritePath3Forward()
        {
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    Assert.IsNull(WritePath(htmlWriter, testNodes, control));
                }

                Assert.AreEqual("titlerootmouse<span>:</span>title1nodemouse<span>:</span>title2nodemouse",
                    writer.ToString(), "The complete path should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing three entries, in a "backward" direction (CurrentToRoot).
        /// </summary>
        [Test]
        public void TestWritePath3Backward()
        {
            using (StringWriter writer = new StringWriter())
            {
                control.PathDirection = PathDirection.CurrentToRoot;
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    Assert.IsNull(WritePath(htmlWriter, testNodes, control));
                }

                Assert.AreEqual("title2nodemouse<span>:</span>title1nodemouse<span>:</span>titlerootmouse",
                    writer.ToString(), "The complete path should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, in a "backward" direction (CurrentToRoot).
        /// </summary>
        [Test]
        public void TestWritePath1Backward()
        {
            using (StringWriter writer = new StringWriter())
            {
                control.PathDirection = PathDirection.RootToCurrent;
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    Assert.IsNull(WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control));
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, in a "forward" direction (RootToCurrent).
        /// </summary>
        [Test]
        public void TestWritePath1Forward()
        {
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    Assert.IsNull(WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control));
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests ending a path with a null parameter.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestEndPathNull()
        {
            EndPath(null);
        }

        /// <summary>
        /// Tests beginning a path with a null parameter.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestBeginPathNull()
        {
            BeginPath(null);
        }

        /// <summary>
        /// Tests writing a path with a null writer, exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestWritePathNull()
        {
            WritePath(null, testNodes, control);
        }

        /// <summary>
        /// Tests writing a separator with a null writer. An exception should be thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestWriteSeparatorNullWriter()
        {
            WriteSeparator(null, new BreadCrumbTrailControl());
        }

        /// <summary>
        /// Tests writing a separator with a null control. An exception should be thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestWriteSeparatorNullControl()
        {
            using (StringWriter writer = new StringWriter())
            {
                try
                {
                    using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                    {
                        WriteSeparator(htmlWriter, null);
                    }
                }
                finally
                {
                    Assert.AreEqual("", writer.ToString(), "Nothing should be written.");
                }
            }
        }

        /// <summary>
        /// Tests writing null entries, also verifying that nothing is written before the exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestWritePathNullNodes()
        {
            using (StringWriter writer = new StringWriter())
            {
                try
                {
                    using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                    {
                        WritePath(htmlWriter, null, control);
                    }
                }
                finally
                {
                    Assert.AreEqual("", writer.ToString(), "Nothing should be written.");
                }
            }
        }

        /// <summary>
        /// Tests writing with a null parent control, also verifying that nothing is written before the exception is
        /// thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestWritePathNullParent()
        {
            using (StringWriter writer = new StringWriter())
            {
                try
                {
                    using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                    {
                        WritePath(htmlWriter, testNodes, null);
                    }
                }
                finally
                {
                    Assert.AreEqual("", writer.ToString(), "Nothing should be written.");
                }
            }
        }

        /// <summary>
        /// Let all the node writings fail, with the normal, silent, error handling.
        /// </summary>
        [Test]
        public void TestWritePathSilentFailNodes()
        {
            using (StringWriter writer = new StringWriter())
            {
                try
                {
                    using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                    {
                        Assert.IsNull(WritePath(htmlWriter, emptyNodes, control));
                    }
                }
                finally
                {
                    Assert.AreEqual("<span>:</span><span>:</span>", writer.ToString(),
                        "Only separators should succeed.");
                }
            }
        }

        /// <summary>
        /// Let all the node writings fail, with the "rendering" error handling (return exceptions).
        /// </summary>
        [Test]
        public void TestWritePathRenderFailNodes()
        {
            control.ErrorDisposition = ErrorDisposition.RenderException;
            using (StringWriter writer = new StringWriter())
            {
                try
                {
                    using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                    {
                        BreadCrumbException[] es = WritePath(htmlWriter, emptyNodes, control);

                        Assert.AreEqual(3, es.Length, "Wrong count of exceptions returned.");
                    }
                }
                finally
                {
                    Assert.AreEqual("<span>:</span><span>:</span>", writer.ToString(),
                        "Only separators should succeed.");
                }
            }
        }

        /// <summary>
        /// Let all the node writings possibly fail, with the "throwing" error handling (return the FIRST exception).
        /// </summary>
        [Test]
        public void TestWritePathThrowFailNodes()
        {
            control.ErrorDisposition = ErrorDisposition.ThrowException;
            using (StringWriter writer = new StringWriter())
            {
                try
                {
                    using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                    {
                        BreadCrumbException[] es = WritePath(htmlWriter, emptyNodes, control);

                        Assert.AreEqual(1, es.Length, "Wrong count of exceptions returned.");
                    }
                }
                finally
                {
                    Assert.AreEqual("", writer.ToString(), "No visible text should succeed.");
                }
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing BeginPath, with silent error handling. Everything still seems to
        /// work.
        /// </summary>
        [Test]
        public void TestWritePathBeginFailSilent()
        {
            failBegin = true;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    Assert.IsNull(WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control));
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing BeginPath, with rendering error handling. Everything still seems to
        /// work, but an exception is returned.
        /// </summary>
        [Test]
        public void TestWritePathBeginFailRender()
        {
            failBegin = true;
            control.ErrorDisposition = ErrorDisposition.RenderException;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    BreadCrumbException[] es = WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control);

                    Assert.AreEqual(1, es.Length, "Wrong count of exceptions returned.");
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing BeginPath, with throwing error handling. The writing is aborted.
        /// </summary>
        [Test]
        public void TestWritePathBeginThrowRender()
        {
            failBegin = true;
            control.ErrorDisposition = ErrorDisposition.ThrowException;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    BreadCrumbException[] es = WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control);

                    Assert.AreEqual(1, es.Length, "Wrong count of exceptions returned.");
                }

                Assert.AreEqual("", writer.ToString(), "Nothing should be written.");
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing EndPath, with silent error handling. Everything still seems to
        /// work.
        /// </summary>
        [Test]
        public void TestWritePathEndFailSilent()
        {
            failEnd = true;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    Assert.IsNull(WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control));
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing EndPath, with rendering error handling. Everything still seems to
        /// work, but an exception is returned.
        /// </summary>
        [Test]
        public void TestWritePathEndFailRender()
        {
            failEnd = true;
            control.ErrorDisposition = ErrorDisposition.RenderException;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    BreadCrumbException[] es = WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control);

                    Assert.AreEqual(1, es.Length, "Wrong count of exceptions returned.");
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing EndPath, with throwing error handling. The writing is aborted.
        /// </summary>
        [Test]
        public void TestWritePathEndThrowRender()
        {
            failEnd = true;
            control.ErrorDisposition = ErrorDisposition.ThrowException;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    BreadCrumbException[] es = WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control);

                    Assert.AreEqual(1, es.Length, "Wrong count of exceptions returned.");
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing Begin/EndPath, with silent error handling. Everything still seems
        /// to work.
        /// </summary>
        [Test]
        public void TestWritePathBeginEndFailSilent()
        {
            failBegin = true;
            failEnd = true;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    Assert.IsNull(WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control));
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }

        /// <summary>
        /// Tests writing one entry, with a failing Begin/EndPath, with rendering error handling. Everything still
        /// seems to work, but an exception is returned.
        /// </summary>
        [Test]
        public void TestWritePathBeginEndFailRender()
        {
            failBegin = true;
            failEnd = true;
            control.ErrorDisposition = ErrorDisposition.RenderException;
            using (StringWriter writer = new StringWriter())
            {
                using (HtmlTextWriter htmlWriter = new HtmlTextWriter(writer))
                {
                    BreadCrumbException[] es = WritePath(htmlWriter, new BreadCrumbNode[1] { testNodes[0] }, control);

                    Assert.AreEqual(2, es.Length, "Wrong count of exceptions returned.");
                }

                Assert.AreEqual("titlerootmouse", writer.ToString(),
                    "The single element should be reported correctly.");
            }
        }
    }
}
